#include "rts_actions.hpp"
#include "nav_marker.hpp"
#include "nav_room.hpp"
#include "rts_states.hpp"
#include "goap/actor_world_state.hpp"
#include "goap/state.hpp"
#include <godot_cpp/core/memory.hpp>
#include <godot_cpp/variant/basis.hpp>
#include <godot_cpp/variant/utility_functions.hpp>
#include <cmath>

MoveToTarget::MoveToTarget()
: Action() {
    this->effects.insert("is_at_target", true);
}

goap::State *MoveToTarget::get_apply_state(goap::ActorWorldState *context) const {
    gd::Node3D *target{gd::Object::cast_to<gd::Node3D>(context->get_world_property("target_node"))};
    if(target == nullptr) {
        gd::UtilityFunctions::push_warning("Failed to get target node of action ", get_static_class());
        return nullptr;
    } else {
        MoveTo *state{this->create_state<MoveTo>()};
        state->target_node = target;
        return state;
    }
}

FireAtTarget::FireAtTarget()
: Action() {
    this->required.insert("can_see_target", true);
    this->effects.insert("is_target_dead", true);
}

goap::State *FireAtTarget::get_apply_state(goap::ActorWorldState *context) const {
    Animate *state{this->create_state<Animate>()};
    state->animation = "fire_weapon";
    return state;
}

FindTarget::FindTarget()
: Action() {
    this->require_state_complete = false;
    this->effects.insert("can_see_target", true);
}

goap::State *FindTarget::get_apply_state(goap::ActorWorldState *context) const {
    gd::Node3D *target{gd::Object::cast_to<gd::Node3D>(context->get_world_property("target_node"))};
    MoveTo *state{this->create_state<MoveTo>()};
    state->target_node = target;
    return state;
}

GetInMeleeRange::GetInMeleeRange()
: Action() {
    this->require_state_complete = false;
    this->required.insert("can_see_target", true);
    this->effects.insert("is_in_melee_range", true);
}

goap::State *GetInMeleeRange::get_apply_state(goap::ActorWorldState *context) const {
    gd::Node3D *target{gd::Object::cast_to<gd::Node3D>(context->get_world_property("target_node"))};
    MoveTo *state{this->create_state<MoveTo>()};
    state->target_node = target;
    return state;
}

MeleeAttack::MeleeAttack()
: Action() {
    this->required.insert("can_see_target", true);
    this->required.insert("is_in_melee_range", true);
    this->effects.insert("is_target_dead", true);
}

goap::State *MeleeAttack::get_apply_state(goap::ActorWorldState *context) const {
    Animate *state{this->create_state<Animate>()};
    state->animation = "melee_attack";
    return state;
}

TankSelfHeal::TankSelfHeal()
: Action() {
    this->required.insert("can_see_target", false);
    this->effects.insert("is_health_safe", true);
}

goap::State *TankSelfHeal::get_apply_state(goap::ActorWorldState *context) const {
    Animate *state{this->create_state<Animate>()};
    state->animation = "self_heal";
    return state;
}

TakeCover::TakeCover()
: Action () {
    this->require_state_complete = false;
    this->effects.insert("can_see_target", false);
}

bool TakeCover::procedural_is_possible(goap::ActorWorldState *context) const {
    gd::Vector3 const global_position{context->get_world_property("parent_global_position")};
    gd::Vector3 const target_position{context->get_world_property("target_global_position")};
    NavRoom *room{NavRoom::get_closest_room(global_position)};
    if(room == nullptr)
        return false;
    for(NavMarker *marker : room->get_markers())
        if(TakeCover::is_marker_cover_from(marker, target_position))
            return true;
    return false;
}

goap::State *TakeCover::get_apply_state(goap::ActorWorldState *context) const {
    MoveTo *state{this->create_state<MoveTo>()};
    gd::Vector3 const global_position{context->get_world_property("parent_global_position")};
    gd::Vector3 const target_position{context->get_world_property("target_global_position")};
    NavRoom *room{NavRoom::get_closest_room(global_position)};
    if(room == nullptr)
        return nullptr;
    float best_score{-INFINITY};
    for(NavMarker *marker : room->get_markers()) {
        gd::Vector3 const marker_position{marker->get_global_position()};
        float const score{100.f - marker_position.distance_to(global_position)};
        if(score > best_score && TakeCover::is_marker_cover_from(marker, target_position)) {
            best_score = score;
            state->target_node = marker;
            gd::UtilityFunctions::print("!!! best cover so far: ", marker->get_path());
        }
    }
    return state;
}

bool TakeCover::is_marker_cover_from(NavMarker *marker, gd::Vector3 const &target) {
    return marker->get_marker_type() == MarkerType::Cover
        && marker->get_global_basis()
                .get_column(2)
                .dot((marker->get_global_position() - target).normalized()) < -0.7f;
}
